#include <ctype.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <strings.h>

#include "types.h"
#include "trtabs.h"

#define DEBUG printf("DEBUG:%s:%d\n",__FILE__,__LINE__);

void	dump16(uchar*) ;
uns2	cchh2tt(uns4);
uns4	tt2cchh(uns2);
int	readblk(void) ;

typedef struct {
	uns2	ccstart;
	uns2	hhstart;
	uns2	ccend;
	uns2	hhend;
	uns2	tracks;
	uns2	tpc;
	}	dsextent ;

typedef struct {
	int4	linect;
	uns2	tt;
	uns1	r;
	char	name[9];
	}	direntry ;

unsigned char * buf;
dsextent * exttab;
int numext;
direntry * de_ar, * deptr=NULL, * deeptr ;
int	foffset=0,boffset=0;
int	debug=0,verbose=0,fl_write=1 ;

static char *UsageMsg=
"USAGE: %s -cvak -r Lrecl -o OutDir [file]\n"
	"\t-o	- output directory; default=.\n"
	"\t-r	- lrecl for F[B] format libraries; forces TEXT MODE\n"
	"\t-k	- force KOI8 text output (default for UNIX)\n"
	"\t-a	- force ALT text output (default for DOS)\n"
	"\t-c	- do not actually write, check format only\n"
	"\t-v	- verobose - more information\n"
	"\t-d	- debugging output\n"
	;

static int	ifd=0 ;

int main(argc,argv)
int 	argc;
char	* argv[1];
{

FILE * outfile=NULL ;
char	c;
char	*imgfilename=NULL ;
uchar	*iptr, *eiptr, *epbptr, *edbptr ;
char	outfilename[256] ;
char	*outstr=NULL ;
int	i,blkl,bytesused;
int	flag_dir ;
char	*outputdir="." ;

uns2	tt,r,de_c ;
uns2	cc,hh,kl,dl ;
int	textmode=0 ;
int	lrecl=0 ;
int	de_datalen ;

#ifdef	unix
char	* TRT_DEFAULT=TRT_DKOI2KOI8 ;
#else
char	* TRT_DEFAULT=TRT_DKOI2ALT ;
#endif

dsextent * extp ;

while ( (c=getopt(argc,argv, "o:?vcr:akd" )) != EOF ) {
        switch(c) {
        case 'o':       outputdir=strdup(optarg); break;
	case 'v':	verbose=1;break;
	case 'd':	verbose=1;debug=1;break;
	case 'c':	fl_write=0;break;
	case 'r':	textmode=1;lrecl=atoi(optarg);break;
	case 'a':	TRT_DEFAULT=TRT_DKOI2ALT;break;
	case 'k':	TRT_DEFAULT=TRT_DKOI2KOI8;break;
        default:        printf(UsageMsg,argv[0]); exit(0);
                }
        }

if (optind <= argc-1) imgfilename=argv[optind] ;

if ( imgfilename && (ifd=open(imgfilename,O_RDONLY))==-1 ) {
	fprintf(stderr,"Can't open '%s': ",imgfilename);
	perror("");
	exit(1);
	}

if ( ( buf = (unsigned char*)malloc(32760) ) == NULL ) {
	fprintf(stderr,"malloc() error\n");
	exit(1) ;
	}

if ( ( de_ar=(direntry*)malloc(2000*sizeof(direntry)) ) == NULL ) {
	fprintf(stderr,"malloc() error\n");
	exit(1) ;
	}

if (textmode) {
	outstr=(char*)malloc(lrecl+1) ;
	*(outstr+lrecl)='\0';
	}


deeptr=de_ar ;

readblk() ; /* ? */

readblk() ;

numext= *(buf+20) - 1 ;
if (verbose) printf("%d extents:\n",numext);
if (verbose) printf("CCCC - HHHH\tCCCC - HHHH   tracks   t/c \n");
exttab=(dsextent*)malloc(numext*sizeof(dsextent)) ;
extp=exttab ;
	for (i=0;i<numext;i++) {
		iptr=buf+24+(16*i) ;
		/*extp=(dsextent*)(exttab+(sizeof(dsextent)*i));*/
		extp->ccstart=	hword370(iptr+6) ;
		extp->hhstart=	hword370(iptr+8) ;
		extp->ccend=	hword370(iptr+10) ;
		extp->hhend=	hword370(iptr+12) ;
		extp->tracks=	hword370(iptr+14) ;

		extp->tpc=( ( extp->ccend != extp->ccstart ) ?
			( extp->tracks-1- (extp->hhend-extp->hhstart) ) /
				( extp->ccend - extp->ccstart ) :
			1000 ) ;

if (verbose) printf("%04hX - %04hX\t%04hX - %04hX    %5d  %3d \n",
  extp->ccstart,extp->hhstart,extp->ccend,extp->hhend,extp->tracks,extp->tpc ) ;

		extp++ ;
		}

flag_dir=1 ;
if (verbose) printf("\n  name   TTTT RR CC  data\n");
/****/
while ( (blkl=readblk() ) > 0 ) {
  epbptr=buf+blkl;
  iptr=buf+8 ;
  while ( iptr < epbptr ) {
    cc=hword370(iptr+4) ;
    hh=hword370(iptr+6) ;
    r=*(iptr+8);
    kl=*(iptr+9) ;
    dl=hword370(iptr+10);
    tt=cchh2tt((((uns4)cc)<<16)+hh) ;

if (debug) printf("+%08X: disk block: %04hX-%04hX %02hX   (tt=%04hX)  kl=%d  dl=%d\n",
			(iptr-buf)+boffset-8,cc,hh,r,tt,kl,dl );
    iptr=iptr+12;	/* skip ph.address */

      edbptr=iptr+kl+dl ;
	if ( edbptr>epbptr ) printf("*** ERROR - invalid 'count' of disk block\n");

	if ( cc==0 && hh==0 && r==0 && flag_dir ) { /*directory */
      		iptr=iptr+8;	/* skip key */
      		bytesused=hword370(iptr);
      		eiptr=iptr+bytesused;
      		iptr=iptr+2;
      		while ( iptr < eiptr ) {
		    if ( *iptr==(unsigned char)0xff ) {
			flag_dir=0 ;
			if (debug) printf(" --- EOF in directory ---\n");
			else if (verbose) printf("\n");
			break ;
			}
		    for (i=0;i<8;i++)
			*(iptr+i)=tolower(TRT_FROMDKOI[*(iptr+i)]);
		    tt=hword370(iptr+8);
		    r=*(iptr+10);
		    de_c=*(iptr+11);
		     deeptr->tt=tt;
		     deeptr->r=r;
		     strncpy(deeptr->name,iptr,8);
		     deeptr->name[8]='\0';
		     deeptr->linect=0;
		    de_datalen=(de_c&0x1f)*2 ;
		     for (i=7;i>0;i--) if (deeptr->name[i]==' ')
						deeptr->name[i]='\0';
		     deeptr++ ;
		     if (verbose) {
			printf("%.8s %04hX %02X %02X ",
					iptr, tt, 0xff & r, 0xff & de_c );
			for (i=0;i<de_datalen;i=i+2)
			   printf(" %02X%02X",
				*(iptr+12+i),*(iptr+12+i+1));
			printf("\n");
			}
		    iptr=iptr+12+de_datalen;
		    }
		iptr=edbptr;
		}

	else if (kl+dl==0) { /* EOF */
		if ( flag_dir ) {
			flag_dir=0 ;
			if (debug) printf(" --- EOF in directory ---\n");
			}
		else {
			if ( outfile ) {
				if (verbose) {
					printf("%6d ",deptr->linect);
					if (textmode) printf("lines\n");
					else printf("bytes\n");
					}
				if (fl_write) fclose(outfile);
				outfile=NULL; }
			if (debug) printf("  --- EOF ---\n");
			deptr=deeptr;
			}
		}

	else /* not dir */ {
		if (debug) printf("  data block\n");
		if ( outfile==NULL ) {
			for (deptr=de_ar;deptr<deeptr;deptr++)
				if ( tt==deptr->tt && r==deptr->r ) break ;
			if (deptr>=deeptr) printf("*** ERROR - no DE for block\n");
			else {
			    if (verbose) printf("%-8.8s -",deptr->name);
			    sprintf(outfilename,"%s/%s",outputdir,deptr->name);

			    if (fl_write) {
				if ( (outfile=fopen(outfilename,"w")) == NULL ){
				    fprintf(stderr,"Can't open %s\n",outfilename);
				    exit(1);
				    }
				}
			    else {outfile=(FILE*)1 ; /*not NULL*/}
			    }
			}

		/* Write block */

		if (textmode) {
      		    while ( iptr < edbptr ) {
			for (i=0;i<lrecl;i++)
				*(outstr+i)=TRT_DEFAULT[*(iptr+i)];
			/*printf("OUT:%s\n",outstr);*/
			if (fl_write) fprintf(outfile,"%s\n",outstr);
			iptr=iptr+lrecl;

			if (iptr>edbptr) {
				printf("*** ERROR *** - wrong lrecl\n");
				iptr=edbptr;
				}

			deptr->linect++;
			}
		    }

		else /*not text*/ {
			if (fl_write) fwrite(iptr,kl+dl,1,outfile);
			iptr=edbptr ;
			deptr->linect+=kl+dl;
			}
		}

    } /*loop inside 1 block*/
  } /*read loop*/

exit(0);

}

/******************************************************************/
int readblk() {
	int bread,blkl,recl ;
	int bo_corr=0,bl_corr=0;

	if (debug) printf("\n+%08X: read block descriptor\n",foffset);
	if ( ((bread=read(ifd,buf,8))<8) ) {
		if (debug) printf("*** END OF FILE ***\n");
		return(0) ; }

/* dump16(buf); fflush(stdout) ; */

	blkl=hword370(buf) ;
	recl=hword370(buf+4) ;
/*** ERROR ***/
if ( blkl==0 || blkl!=recl+4 || *(short*)(buf+2)!=0 || *(short*)(buf+6)!=0 ) {
	printf("*** ERROR - bad block descriptor at +%08X :\n",foffset);
	dump16(buf);
	if ( deptr && deptr->name ) printf("Member=%s\n",deptr->name);
fflush(stdout);
	}

	foffset=foffset+8;
	blkl=blkl-bl_corr;

	if (debug) printf("+%08X read blkl=%d\n",foffset,blkl);
	if ( ((bread=read(ifd,buf+8+bo_corr,blkl-8))<blkl-8) ) return(0) ;
	boffset=foffset ;
	foffset=foffset+blkl-8;
	return(blkl) ;
	}

uns2 cchh2tt( uns4 cchh ) {
	int i,cc,hh,tt=0 ;
	dsextent * extp = exttab ;
	if ( cchh == 0 ) return(0) ;
	hh=cchh&0xffff; cc=cchh>>16;

	for (i=0;i<numext;i++) {
/*printf("testing : %04hX - %02hX\t%04hX - %02hX : %05d \n",
 extp->ccstart, extp->hhstart, extp->ccend, extp->hhend, extp->tracks );*/
	  if ( ( cc>extp->ccstart || (cc==extp->ccstart && hh>=extp->hhstart) )
	      && ( cc<extp->ccend || (cc==extp->ccend && hh<=extp->hhend) ) ) {
		    return(tt+(cc-extp->ccstart)*extp->tpc+hh-extp->hhstart);
		    }
		tt=tt+extp->tracks ;
		extp++ ;
		}
	printf("*** ERROR - cchh out of extent\n");
	return(-1) ;
	}

uns4 tt2cchh( uns2 tt ) {
	int cc,hh ;
	dsextent * extp = exttab ;
	/* numext ?*/
	while ( tt > extp->tracks ) {
		tt=tt-extp->tracks ;
		extp++ ;
		}
	hh = ( extp->hhstart + tt ) % extp->tpc ;
	cc = ( extp->hhstart + tt ) / extp->tpc + extp->ccstart ;
	return((((uns4)cc)<<16)+hh) ;
	}
